"""
Utility functions for logging, staging data results for external
access (S3 pre-signed URL), decrypting data using a shared secret, and
operating on message queues.

This module relies heavily on environment variables to know
which endpoints to use and how to authenticate to them as follows:

Required when receiving an encrypted user access token in the message (Always!):
    SHARED_SECRET_KEY:  The 32-byte shared encryption / decryption key to decrypt
                        the access token in the Harmony operation.

Required when reading from or staging to S3:
    AWS_DEFAULT_REGION: The AWS region in which the S3 client is operating (default: "us-west-2")

Required when staging to S3 and not using the Harmony-provided stagingLocation prefix:
    STAGING_BUCKET: The bucket where staged files should be placed
    STAGING_PATH: The base path under which staged files should be placed

Required when using HTTPS, allowing Earthdata Login auth:
    OAUTH_HOST:         The Earthdata Login (EDL) environment to connect to
    OAUTH_CLIENT_ID:    The EDL application client id used to acquire an EDL shared access token
    OAUTH_UID:          The EDL application UID used to acquire an EDL shared access token
    OAUTH_PASSWORD:     The EDL application password used to acquire an EDL shared access token
    OAUTH_REDIRECT_URI: A valid redirect URI for the EDL application (NOTE: the redirect URI is
                        not followed or used; it does need to be in the app's redirect URI list)

Always provided by newer versions of the Harmony frontend:
    USER_AGENT:     The Harmony user agent string. E.g. harmony/0.0.0 (harmony-sit)

Optional, if support is needed for downloading data from an endpoint that is not
EDL-share-token aware:

    FALLBACK_AUTHN_ENABLED: Whether to try downloading with the EDL_* credentials.
    EDL_USERNAME:           An valid EDL user entity username.
    EDL_PASSWORD:           The password belonging to EDL_USERNAME.

Optional when reading from or staging to S3:
    USE_LOCALSTACK:  'true' if the S3 client should connect to a LocalStack instance instead of
                     Amazon S3 (for testing)
    LOCALSTACK_HOST: The hostname of the Localstack to connect to if `USE_LOCALSTACK`.
    BACKEND_HOST:    The hostname of the Harmony backend. Deprecated / unused by this package.

Optional:
    APP_NAME:             A name for the service that will appear in log entries.
    ENV:                  The application environment. One of: dev, test. Used for local development.
    TEXT_LOGGER:          Whether to log in plaintext or JSON. Default: True (plaintext).
    HEALTH_CHECK_PATH:    The filesystem path that should be `touch`ed to indicate the service is
                          alive.
    MAX_DOWNLOAD_RETRIES: Number of times to retry HTTP download calls that fail due to transient errors.
"""

from base64 import b64decode
from collections import namedtuple
from functools import lru_cache
import hashlib
import logging
from pathlib import Path, PurePath
from os import environ, path
import sys
import re
from urllib import parse

from nacl.secret import SecretBox

from harmony import aws
from harmony import http
# The following imports are for backwards-compatibility for services
# which import them from `harmony.util`. Though they are not used in
# this module, importing them here allows applications to work without
# modifications.
from harmony.exceptions import (HarmonyException, CanceledException, ForbiddenException)  # noqa: F401
from harmony.logging import build_logger
from harmony.version import get_version


DEFAULT_SHARED_SECRET_KEY = '_THIS_IS_MY_32_CHARS_SECRET_KEY_'


Config = namedtuple(
    'Config', [
        'app_name',
        'oauth_host',
        'oauth_client_id',
        'oauth_uid',
        'oauth_password',
        'oauth_redirect_uri',
        'fallback_authn_enabled',
        'edl_username',
        'edl_password',
        'use_localstack',
        'backend_host',
        'localstack_host',
        'aws_default_region',
        'staging_path',
        'staging_bucket',
        'env',
        'text_logger',
        'health_check_path',
        'shared_secret_key',
        'user_agent',
        'max_download_retries'
    ])


def _validated_config(config):
    """Validates that the given Config has values for all required
    variables and returns it if so. Raises an Exception if invalid.
    """
    required = [
        'shared_secret_key',
        'oauth_client_id',
        'oauth_uid',
        'oauth_password',
        'oauth_redirect_uri',
        'staging_path',
        'staging_bucket',
        'max_download_retries'
    ]

    unset = [var.upper() for var in required if getattr(config, var) is None]

    # Conditionally required
    if config.fallback_authn_enabled and getattr(config, 'edl_username') is None:
        unset.append("EDL_USERNAME")
    if config.fallback_authn_enabled and getattr(config, 'edl_password') is None:
        unset.append("EDL_PASSWORD")

    if len(unset) > 0:
        msg = f"Required environment variables are not set: {', '.join(unset)}"
        raise Exception(msg)

    # Warnings
    if config.shared_secret_key == DEFAULT_SHARED_SECRET_KEY:
        logging.warning("The SHARED_SECRET_KEY has not been set. Currently set to its default (unsecure) value.")

    logging.info(config)

    return config


@lru_cache(maxsize=128)
def config(validate=True):
    """
    Returns the Config object with all parameters set to values that were set in the
    process' environment (as environment variables), or to their default values if not
    set.

    Parameters
    ----------
    validate : bool
        Whether to validate the config before returning it. Useful to disable when
        running unit tests.

    Returns
    -------
    harmony.util.Config
        The configuration values for this runtime environment.
    """
    def str_envvar(name: str, default: str) -> str:
        value = environ.get(name, default)
        return value.strip('\"') if value is not None else None

    def bool_envvar(name: str, default: bool) -> bool:
        value = environ.get(name)
        return str.lower(value) == 'true' if value is not None else default

    def int_envvar(name: str, default: int) -> int:
        value = environ.get(name)
        return int(value) if value is not None else default

    oauth_redirect_uri = str_envvar('OAUTH_REDIRECT_URI', None)
    if oauth_redirect_uri is not None:
        oauth_redirect_uri = parse.quote(oauth_redirect_uri)
    backend_host = str_envvar('BACKEND_HOST', 'localhost')
    localstack_host = str_envvar('LOCALSTACK_HOST', backend_host)

    config = Config(
        app_name=str_envvar('APP_NAME', sys.argv[0]),
        oauth_host=str_envvar('OAUTH_HOST', 'https://uat.urs.earthdata.nasa.gov'),
        oauth_client_id=str_envvar('OAUTH_CLIENT_ID', None),
        oauth_uid=str_envvar('OAUTH_UID', None),
        oauth_password=str_envvar('OAUTH_PASSWORD', None),
        oauth_redirect_uri=oauth_redirect_uri,
        fallback_authn_enabled=bool_envvar('FALLBACK_AUTHN_ENABLED', False),
        edl_username=str_envvar('EDL_USERNAME', None),
        edl_password=str_envvar('EDL_PASSWORD', None),
        use_localstack=bool_envvar('USE_LOCALSTACK', False),
        backend_host=backend_host,
        localstack_host=localstack_host,
        aws_default_region=str_envvar('AWS_DEFAULT_REGION', 'us-west-2'),
        staging_path=str_envvar('STAGING_PATH', None),
        staging_bucket=str_envvar('STAGING_BUCKET', None),
        env=str_envvar('ENV', ''),
        text_logger=bool_envvar('TEXT_LOGGER', False),
        health_check_path=str_envvar('HEALTH_CHECK_PATH', '/tmp/health.txt'),
        shared_secret_key=str_envvar('SHARED_SECRET_KEY', DEFAULT_SHARED_SECRET_KEY),
        user_agent=str_envvar('USER_AGENT', 'harmony (unknown version)'),
        max_download_retries=int_envvar('MAX_DOWNLOAD_RETRIES', 0)
    )

    if validate:
        return _validated_config(config)
    else:
        return config


def _build_full_user_agent(config) -> str:
    """
    Builds a user-agent string that can be passed on to aws or http clients.
    The user agent may consist of a user agent defined by an env variable passed
    by newer versions of Harmony, a user agent for this service lib, and an optional user
    agent that can be provided by users of this lib.

    Parameters
    ----------
    config : harmony.util.Config
        The configuration values for this runtime environment.

    Returns
    -------
    string
        A user agent string.
    """
    harmony_user_agent = config.user_agent
    app_name = config.app_name
    lib_user_agent = f'harmony-service-lib/{get_version()}'
    full_user_agent = f'{harmony_user_agent} {lib_user_agent}'
    if app_name is not None:
        full_user_agent += f' ({app_name})'
    return full_user_agent


def _is_file_url(url: str) -> bool:
    return url is not None and url.startswith('file://')


def _url_as_filename(url: str) -> str:
    """Return a version of the url optimized for local development.

    If the url is a `file://` url, it will return the remaining part
    of the url so it can be used as a local file path. For example,
    'file:///logs/example.txt' will be converted to
    '/logs/example.txt'.

    Parameters
    ----------
    url: str The url to check and optaimize.
    Returns
    -------
    str: The url converted to a filename.

    """
    return url.replace('file://', '')


def _filename(directory_path: str, url: str) -> Path:
    """Constructs a filename from the url using the specified directory
    as its path. The constructed filename will be a sha256 hash
    (converted to a hex digest) of the url, and the file's extension
    will be the same as that of the filename in the url.

    Parameters
    ----------
    directory_path : str
        The url to use when constructing the filename and extension.
    Returns
    -------

    """
    return Path(
        directory_path,
        hashlib.sha256(url.encode('utf-8')).hexdigest()
    ).with_suffix(PurePath(parse.urlparse(url).path).suffix)


def download(url, destination_dir, logger=None, access_token=None, data=None, cfg=None):
    """
    Downloads the given URL to the given destination directory, using the basename of the URL
    as the filename in the destination directory.  Supports http://, https:// and s3:// schemes.
    When using the s3:// scheme, will run against us-west-2 unless the "AWS_DEFAULT_REGION"
    environment variable is set.

    When using http:// or https:// schemes, the access_token will be used for authentication.

    Parameters
    ----------
    url : string
        The URL to fetch
    destination_dir : string
        The directory in which to place the downloaded file
    logger : Logger
        A logger to which the function will write, if provided
    access_token :
        The Earthdata Login token of the caller to use for downloads
    data : dict or Tuple[str, str]
        Optional parameter for additional data to
        send to the server when making a HTTP POST request through
        urllib.get.urlopen. These data will be URL encoded to a query string
        containing a series of `key=value` pairs, separated by ampersands. If
        None (the default), urllib.get.urlopen will use the  GET
        method.
    cfg : harmony.util.Config
        The configuration values for this runtime environment.

    Returns
    -------
    destination : string
      The filename, including directory, of the downloaded file
    """
    if cfg is None:
        cfg = config()
    if logger is None:
        logger = build_logger(cfg)

    if _is_file_url(url):
        return _url_as_filename(url)

    source = http.localhost_url(url, cfg.localstack_host)

    destination_path = _filename(destination_dir, url)
    if destination_path.exists():
        return str(destination_path)
    destination_path = str(destination_path)

    full_user_agt = _build_full_user_agent(cfg)

    with open(destination_path, 'wb') as destination_file:
        if aws.is_s3(source):
            aws.download(cfg, source, destination_file, full_user_agt)
        elif http.is_http(source):
            http.download(cfg, source, access_token, data, destination_file, full_user_agt)
        else:
            msg = f'Unable to download a url of unknown type: {url}'
            logger.error(msg)
            raise Exception(msg)

    return destination_path


def stage(local_filename, remote_filename, mime, logger=None, location=None, cfg=None):
    """
    Stages the given local filename, including directory path, to an S3 location with the given
    filename and mime-type

    Requires the following environment variables:
        AWS_DEFAULT_REGION: The AWS region in which the S3 client is operating

    Parameters
    ----------
    local_filename : string
        A path and filename to the local file that should be staged
    remote_filename : string
        The basename to give to the remote file
    mime : string
        The mime type to apply to the staged file for use when it is served, e.g. "application/x-netcdf4"
    location : string
        The S3 prefix URL under which to place the output file.  If not provided, STAGING_BUCKET and
        STAGING_PATH must be set in the environment
    logger : logging
        The logger to use
    cfg : harmony.util.Config
        The configuration values for this runtime environment.

    Returns
    -------
    url : string
        An s3:// URL to the staged file
    """
    # The implementation of this function has been moved to the
    # harmony.aws module.
    if cfg is None:
        cfg = config()
    if logger is None:
        logger = build_logger(cfg)

    return aws.stage(cfg, local_filename, remote_filename, mime, logger, location)


def receive_messages(queue_url, visibility_timeout_s=600, logger=None, cfg=None):
    """
    Generates successive messages from reading the queue.  The caller
    is responsible for deleting or returning each message to the queue

    Parameters
    ----------
    queue_url : string
        The URL of the queue to receive messages on
    visibility_timeout_s : int
        The number of seconds to wait for a received message to be deleted
        before it is returned to the queue
    cfg : harmony.util.Config
        The configuration values for this runtime environment.

    Yields
    ------
    receiptHandle, body : string, string
        A tuple of the receipt handle, used to delete or update messages,
        and the contents of the message
    """
    # The implementation of this function has been moved to the
    # harmony.aws module.
    if cfg is None:
        cfg = config()
    if logger is None:
        logger = build_logger(cfg)

    touch_health_check_file(cfg.health_check_path)
    return aws.receive_messages(cfg, queue_url, visibility_timeout_s, logger)


def delete_message(queue_url, receipt_handle, cfg=None):
    """
    Deletes the message with the given receipt handle from the provided queue URL,
    indicating successful processing

    Parameters
    ----------
    queue_url : string
        The queue from which the message originated
    receipt_handle : string
        The receipt handle of the message, as yielded by `receive_messages`
    cfg : harmony.util.Config
        The configuration values for this runtime environment.
    """
    # The implementation of this function has been moved to the
    # harmony.aws module.
    if cfg is None:
        cfg = config()
    return aws.delete_message(cfg, queue_url, receipt_handle)


def change_message_visibility(queue_url, receipt_handle, visibility_timeout_s, cfg=None):
    """
    Updates the message visibility timeout of the message with the given receipt handle

    Parameters
    ----------
    queue_url : string
        The queue from which the message originated
    receipt_handle : string
        The receipt handle of the message, as yielded by `receive_messages`
    visibility_timeout_s : int
        The number of additional seconds to wait for a received message to be deleted
        before it is returned to the queue
    cfg : harmony.util.Config
        The configuration values for this runtime environment.
    """
    # The implementation of this function has been moved to the
    # harmony.aws module.
    if cfg is None:
        cfg = config()
    return aws.change_message_visibility(cfg, queue_url, receipt_handle, visibility_timeout_s)


def touch_health_check_file(health_check_path):
    """
    Updates the mtime of the health check file.
    """
    Path(health_check_path).touch()


def create_decrypter(key=b'_THIS_IS_MY_32_CHARS_SECRET_KEY_'):
    """Creates a function that will decrypt cyphertext using a shared secret
    (symmetric) 32-byte key.

    The returned decrypter function has type signature: str -> str.
    """
    box = SecretBox(key)

    def decrypter(encrypted_msg_str):
        """Decrypt encrypted text using the shared secret (symmetric) key
        in the function's closure."""

        parts = encrypted_msg_str.split(':')
        nonce = b64decode(parts[0])
        ciphertext = b64decode(parts[1])

        return box.decrypt(ciphertext, nonce).decode('utf-8')

    return decrypter


def nop_decrypter(ciphertext):
    """An identity decrypter function. A NOP: it returns the ciphertext
    as-is. Its other responsibility is to have nothing to do with
    crypto-currency transactions in exactly the same way that it has
    nothing to do with quantum computing.
    """
    return ciphertext


def generate_output_filename(filename, ext=None, variable_subset=None, is_regridded=False, is_subsetted=False):
    """
    Return an output filename for the given granules according to our naming conventions:
    {original filename without suffix}(_{single var})?(_regridded)?(_subsetted)?.<ext>

    Parameters
    ----------
        granule : message.Granule
            The source granule for the output file
        ext: string, optional
            The destination file extension (default: original extension)
        variable_subset : string[], optional
            When variable subsetting, a list of all variables that have been subset
        is_regridded : bool, optional
            True if a regridding operation has been performed (default: False)
        is_subsetted : bool, optional
            True if a subsetting operation has been performed (default: False)

    Returns
    -------
        string
            The output filename
    """
    url = filename
    # Get everything between the last non-trailing '/' before the query and the first '?'
    # Do this instead of using a URL parser, because our URLs are not complex in practice and
    # it is useful to allow relative file paths to work for local testing.
    original_filename = url.split('?')[0].rstrip('/').split('/')[-1]
    decoded_original_filename = parse.unquote(original_filename)
    (original_basename, original_ext) = path.splitext(decoded_original_filename)
    if ext is None:
        ext = original_ext

    if not ext.startswith('.'):
        ext = '.' + ext

    suffixes = []
    if variable_subset and len(variable_subset) == 1:
        var = variable_subset[0]
        if hasattr(var, 'name'):
            var = var.name
        suffixes.append('_' + var)
    if is_regridded:
        suffixes.append('_regridded')
    if is_subsetted:
        suffixes.append('_subsetted')
    suffixes.append(ext)

    result = original_basename
    # Iterate suffixes in reverse, removing them from the result if they're at the end of the string
    # This supports the case of chaining where one service regrids and another subsets but we don't
    # want names to get mangled
    for suffix in suffixes[::-1]:
        if result.endswith(suffix):
            result = result[:-len(suffix)]

    result += "".join(suffixes)

    # replace special chars that may have been encoded or present in variable_subset
    result = re.sub('\\/|:', '_', result)

    # runs of underscores are replaced with single underscore
    result = re.sub(r'_{2,}', '_', result)

    # leading or trailing underscores are removed
    result = re.sub(r'^_+|_+$', '', result)

    # underscores before or after periods are removed
    result = re.sub(r'_{0,}\._{0,}', '.', result)

    return result


def bbox_to_geometry(bbox):
    '''
    Creates a GeoJSON geometry given a GeoJSON BBox, accounting for antimeridian

    Parameters
    ----------
    bbox : float[4]
        the bounding box to create a geometry from

    Returns
    -------
    dict
        a GeoJSON Polygon or MultiPolygon representation of the input bbox
    '''
    if not bbox:
        return None
    west, south, east, north = bbox[0:4]
    if west > east:
        return {
            'type': 'MultiPolygon',
            'coordinates': [
                [[
                    [-180, south],
                    [-180, north],
                    [east, north],
                    [east, south],
                    [-180, south]
                ]],
                [[
                    [west, south],
                    [west, north],
                    [180, north],
                    [180, south],
                    [west, south]
                ]]
            ]
        }
    return {
        'type': 'Polygon',
        'coordinates': [[
            [west, south],
            [west, north],
            [east, north],
            [east, south],
            [west, south]
        ]],
    }



"""
Tests deprecated methods and non-STAC invocation styles on BaseHarmonyAdapter
"""

import unittest
from unittest.mock import patch, MagicMock
from tempfile import NamedTemporaryFile, mkdtemp
import os
from os import path, remove
import pathlib
from shutil import rmtree
from urllib.error import HTTPError

from harmony.adapter import BaseHarmonyAdapter
from harmony.message import Message, Granule, Variable, Temporal
import harmony.util
from harmony.exceptions import CanceledException
from .example_messages import minimal_message, full_message


# BaseHarmonyAdapter is abstract, so tests need a minimal concrete class
class AdapterTester(BaseHarmonyAdapter):
    def __init__(self, message_str, config):
        super().__init__(Message(message_str), config=config)

    def invoke(self):
        pass


class MockHTTPError(HTTPError):
    def __init__(self, url='http://example.com', code=409, msg='Harmony canceled request', hdrs=[], fp=None):
        super().__init__(url, code, msg, hdrs, fp)

    def read(self):
        return MagicMock(return_value='request body')


class TestBaseHarmonyAdapter(unittest.TestCase):
    def setUp(self):
        self.config = harmony.util.config(validate=False)

    def test_cleanup_deletes_temporary_file_paths(self):
        adapter = AdapterTester(minimal_message, self.config)
        f = NamedTemporaryFile(delete=False)
        try:
            f.close()
            adapter.temp_paths += [f.name]
            self.assertTrue(path.exists(f.name))

            adapter.cleanup()

            self.assertFalse(path.exists(f.name))
        finally:
            if path.exists(f.name):
                remove(f.name)

    def test_cleanup_deletes_temporary_directory_paths(self):
        adapter = AdapterTester(minimal_message, self.config)
        dirname = mkdtemp()
        try:
            adapter.temp_paths += [dirname]
            self.assertTrue(path.exists(dirname))

            adapter.cleanup()

            self.assertFalse(path.exists(dirname))
        finally:
            if path.exists(dirname):
                rmtree(dirname)

    def test_download_granules_fetches_remote_granules_and_stores_their_path(self):
        adapter = AdapterTester(full_message, self.config)
        try:
            adapter.download_granules()
            granules = adapter.message.granules
            self.assertEqual(granules[0].local_filename, 'example/example_granule_1.txt')
            self.assertEqual(granules[1].local_filename, 'example/example_granule_2.txt')
            self.assertEqual(granules[2].local_filename, 'example/example_granule_3.txt')
            self.assertEqual(granules[3].local_filename, 'example/example_granule_4.txt')
        finally:
            adapter.cleanup()

    def test_download_granules_adds_granule_temp_dir_to_temp_paths(self):
        adapter = AdapterTester(full_message, self.config)
        try:
            self.assertEqual(len(adapter.temp_paths), 0)
            adapter.download_granules()
            self.assertEqual(len(adapter.temp_paths), 1)
        finally:
            adapter.cleanup()

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_completed_with_error_when_no_callback_has_been_made_it_posts_the_error(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.completed_with_error('ohai there')
        _callback_post.assert_called_with('/response?error=ohai%20there')

    def test_completed_with_error_when_a_callback_has_been_made_it_throws_an_exception(self):
        adapter = AdapterTester(full_message, self.config)
        adapter.completed_with_error('ohai there')
        self.assertRaises(Exception, adapter.completed_with_error, 'ohai there again')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_completed_with_redirect_when_no_callback_has_been_made_it_posts_the_redirect(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.completed_with_redirect('https://example.com')
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com'
                                          '&item[type]=image/tiff'
                                          '&status=successful')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_completed_with_redirect_accepts_item_metadata(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        granule = adapter.message.sources[0].granules[0]
        temporal = Temporal(start='2011-11-11T11:11:11Z', end='2011-11-11T11:11:12Z')
        adapter.completed_with_redirect('https://example.com', title='hi', mime='image/tiff', source_granule=granule,
                                        temporal=temporal, bbox=[1.1, 2.2, 3.3, 4.4])
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com'
                                          '&item[type]=image/tiff'
                                          '&item[temporal]=2001-01-01T01%3A01%3A01Z%2C2002-02-02T02%3A02%3A02Z'
                                          '&item[bbox]=-1%2C-2%2C3%2C4'
                                          '&status=successful')

    def test_completed_with_redirect_when_a_callback_has_been_made_it_throws_an_exception(self):
        adapter = AdapterTester(full_message, self.config)
        adapter.completed_with_redirect('https://example.com/1')
        self.assertRaises(Exception, adapter.completed_with_error, 'https://example.com/2')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    @patch.object(harmony.util, 'stage', return_value='https://example.com/out')
    def test_completed_with_local_file_stages_the_local_file_and_redirects_to_it(self, stage, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.completed_with_local_file('tmp/output.tif', remote_filename='out.tif')
        stage.assert_called_with('tmp/output.tif',
                                 'out.tif',
                                 'image/tiff',
                                 location='s3://example-bucket/public/some-org/some-service/some-uuid/',
                                 logger=adapter.logger,
                                 cfg=self.config)
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com/out'
                                          '&item[type]=image/tiff'
                                          '&status=successful')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    @patch.object(harmony.util, 'stage', return_value='https://example.com/out')
    def test_completed_with_local_file_uses_granule_file_naming(self, stage, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        granule = adapter.message.sources[0].granules[0]
        adapter.completed_with_local_file('tmp/output.tif', source_granule=granule,
                                          is_variable_subset=True, is_regridded=True, is_subsetted=True)
        stage.assert_called_with('tmp/output.tif',
                                 'example_granule_1_ExampleVar1_regridded_subsetted.tif',
                                 'image/tiff',
                                 location='s3://example-bucket/public/some-org/some-service/some-uuid/',
                                 logger=adapter.logger,
                                 cfg=self.config)
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com/out'
                                          '&item[type]=image/tiff'
                                          '&item[temporal]=2001-01-01T01%3A01%3A01Z%2C2002-02-02T02%3A02%3A02Z'
                                          '&item[bbox]=-1%2C-2%2C3%2C4'
                                          '&status=successful')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_async_add_url_partial_result_for_async_incomplete_requests_posts_the_url(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        adapter.async_add_url_partial_result('https://example.com')
        _callback_post.assert_called_with('/response?item[href]=https%3A//example.com&item[type]=image/tiff')

    def test_async_add_url_partial_result_for_sync_requests_throws_an_error(self):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = True
        self.assertRaises(Exception, adapter.async_add_url_partial_result, 'https://example.com/2')

    def test_async_add_url_partial_result_for_complete_requests_throws_an_error(self):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        adapter.completed_with_redirect('https://example.com/1')
        self.assertRaises(Exception, adapter.async_add_url_partial_result, 'https://example.com/2')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_async_add_url_partial_result_supplies_bbox_and_temporal_when_provided(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        temporal = Temporal(start='2011-11-11T11:11:11Z', end='2011-11-11T11:11:12Z')
        adapter.async_add_url_partial_result('https://example.com', temporal=temporal, bbox=[1.1, 2.2, 3.3, 4.4])
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com'
                                          '&item[type]=image/tiff'
                                          '&item[temporal]=2011-11-11T11%3A11%3A11Z%2C2011-11-11T11%3A11%3A12Z'
                                          '&item[bbox]=1.1%2C2.2%2C3.3%2C4.4')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_async_add_url_partial_result_uses_granule_bbox_and_temporal_when_provided(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        granule = adapter.message.sources[0].granules[0]
        adapter.async_add_url_partial_result('https://example.com', source_granule=granule)
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com'
                                          '&item[type]=image/tiff'
                                          '&item[temporal]=2001-01-01T01%3A01%3A01Z%2C2002-02-02T02%3A02%3A02Z'
                                          '&item[bbox]=-1%2C-2%2C3%2C4')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_async_add_url_partial_result_prefers_explicit_bbox_and_temporal_over_granule_values(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        granule = adapter.message.sources[0].granules[0]
        temporal = Temporal(start='2011-11-11T11:11:11Z', end='2011-11-11T11:11:12Z')
        adapter.async_add_url_partial_result('https://example.com', source_granule=granule,
                                             temporal=temporal, bbox=[1.1, 2.2, 3.3, 4.4])
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com'
                                          '&item[type]=image/tiff'
                                          '&item[temporal]=2011-11-11T11%3A11%3A11Z%2C2011-11-11T11%3A11%3A12Z'
                                          '&item[bbox]=1.1%2C2.2%2C3.3%2C4.4')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    def test_async_completed_successfully_for_async_incomplete_requests_posts_the_completion_status(self, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        adapter.async_completed_successfully()
        _callback_post.assert_called_with('/response?status=successful')

    def test_async_completed_successfully_for_sync_requests_throws_an_error(self):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = True
        self.assertRaises(Exception, adapter.async_completed_successfully)

    def test_async_completed_successfully_for_complete_requests_throws_an_error(self):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        adapter.async_completed_successfully()
        self.assertRaises(Exception, adapter.async_completed_successfully)

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    @patch.object(harmony.util, 'stage', return_value='https://example.com/out')
    def test_async_add_local_file_partial_result_stages_the_local_file_and_updates_progress(self, stage, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        adapter.async_add_local_file_partial_result('tmp/output.tif', remote_filename='out.tif', title='my file',
                                                    progress=50)
        stage.assert_called_with('tmp/output.tif',
                                 'out.tif',
                                 'image/tiff',
                                 location='s3://example-bucket/public/some-org/some-service/some-uuid/',
                                 logger=adapter.logger,
                                 cfg=self.config)
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com/out'
                                          '&item[type]=image/tiff'
                                          '&item[title]=my%20file'
                                          '&progress=50')

    @patch.object(BaseHarmonyAdapter, '_callback_post')
    @patch.object(harmony.util, 'stage', return_value='https://example.com/out')
    def test_async_add_local_file_partial_result_uses_granule_file_naming(self, stage, _callback_post):
        adapter = AdapterTester(full_message, self.config)
        adapter.message.isSynchronous = False
        granule = adapter.message.sources[0].granules[0]
        adapter.async_add_local_file_partial_result('tmp/output.tif', source_granule=granule,
                                                    is_variable_subset=True, is_regridded=True, is_subsetted=True,
                                                    title='my file', progress=50)
        stage.assert_called_with('tmp/output.tif',
                                 'example_granule_1_ExampleVar1_regridded_subsetted.tif',
                                 'image/tiff',
                                 location='s3://example-bucket/public/some-org/some-service/some-uuid/',
                                 logger=adapter.logger,
                                 cfg=self.config)
        _callback_post.assert_called_with('/response'
                                          '?item[href]=https%3A//example.com/out'
                                          '&item[type]=image/tiff'
                                          '&item[title]=my%20file'
                                          '&item[temporal]=2001-01-01T01%3A01%3A01Z%2C2002-02-02T02%3A02%3A02Z'
                                          '&item[bbox]=-1%2C-2%2C3%2C4'
                                          '&progress=50')

    def test_filename_for_granule(self):
        adapter = AdapterTester(minimal_message, self.config)
        granule = Granule({'url': 'https://example.com/fake-path/abc.123.nc/?query=true'})
        ext = 'zarr'

        # Basic cases
        self.assertEqual(adapter.filename_for_granule(granule, ext), 'abc.123.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_subsetted=True), 'abc.123_subsetted.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_regridded=True), 'abc.123_regridded.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_subsetted=True, is_regridded=True),
                         'abc.123_regridded_subsetted.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True, is_subsetted=True,
                                                      is_regridded=True),
                         'abc.123_regridded_subsetted.zarr')

        # Variable name contains full path with '/' ('/' replaced with '_')
        granule.variables.append(Variable({'name': '/path/to/VarB'}))
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True, is_subsetted=True,
                                                      is_regridded=True),
                         'abc.123__path_to_VarB_regridded_subsetted.zarr')
        granule.variables.pop()

        # Single variable cases
        granule.variables.append(Variable({'name': 'VarA'}))
        self.assertEqual(adapter.filename_for_granule(granule, ext), 'abc.123.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_subsetted=True, is_regridded=True),
                         'abc.123_regridded_subsetted.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True), 'abc.123_VarA.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True, is_subsetted=True,
                                                      is_regridded=True),
                         'abc.123_VarA_regridded_subsetted.zarr')

        # Multiple variable cases (no variable name in suffix)
        granule.variables.append(Variable({'name': 'VarB'}))
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_subsetted=True, is_regridded=True),
                         'abc.123_regridded_subsetted.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True, is_subsetted=True,
                                                      is_regridded=True),
                         'abc.123_regridded_subsetted.zarr')
        granule.variables.pop()

        # URL already containing a suffix
        granule.url = 'https://example.com/fake-path/abc.123_regridded.zarr'
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_subsetted=True),
                         'abc.123_regridded_subsetted.zarr')
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True, is_subsetted=True,
                                                      is_regridded=True),
                         'abc.123_VarA_regridded_subsetted.zarr')

        # URL already containing all suffixes
        granule.url = 'https://example.com/fake-path/abc.123_VarA_regridded_subsetted.zarr'
        self.assertEqual(adapter.filename_for_granule(granule, ext, is_variable_subset=True, is_subsetted=True,
                                                      is_regridded=True),
                         'abc.123_VarA_regridded_subsetted.zarr')

    @patch('urllib.request.urlopen')
    @patch.dict(os.environ, {'ENV': 'not_test_we_swear'})
    def test_cancel_request(self, urlopen):
        adapter = AdapterTester(minimal_message, self.config)
        urlopen.side_effect = MockHTTPError(url='http://example.com', code=409, msg='Harmony canceled request', hdrs=[], fp=None)
        self.assertRaises(CanceledException, adapter.async_add_url_partial_result, 'https://example.com/2')
        self.assertTrue(adapter.is_canceled)
        self.assertTrue(adapter.is_complete)
        self.assertEqual(adapter.is_failed, 1)


class TestCallbackPostHealthUpdate(unittest.TestCase):
    def setUp(self):
        self.config = harmony.util.config(validate=False)

    @patch.object(pathlib.Path, '__new__')
    def test_callback_post_updates_health_check_file(self, mock_path):
        adapter = AdapterTester(full_message, self.config)
        adapter.completed_with_error('ohai there')
        mock_path.return_value.touch.assert_called()



"""
======
cli.py
======

Parses CLI arguments provided by Harmony and invokes the subsetter accordingly
"""

import json
import logging
from os import path, makedirs
import datetime

from pystac import Catalog, CatalogType

from harmony.exceptions import CanceledException, HarmonyException
from harmony.message import Message
from harmony.logging import setup_stdout_log_formatting, build_logger
from harmony.util import (receive_messages, delete_message, change_message_visibility,
                          config, create_decrypter)
from harmony.version import get_version
from harmony.aws import is_s3, write_s3


def setup_cli(parser):
    """
    Adds Harmony arguments to the CLI being parsed by the provided parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser being used to parse CLI arguments
    """
    parser.add_argument('--harmony-action',
                        choices=['invoke', 'start'],
                        help=('the action Harmony needs to perform, "invoke" to run once and quit, '
                              '"start" to listen to a queue'))
    parser.add_argument('--harmony-input',
                        help=('the input data for the action provided by Harmony, required for '
                              '--harmony-action=invoke'))
    parser.add_argument('--harmony-sources',
                        help=('file path that contains a STAC catalog with items and metadata to '
                              'be processed by the service.  Required for non-deprecated '
                              'invocations '))
    parser.add_argument('--harmony-metadata-dir',
                        help=('file path where output metadata should be written. The resulting '
                              'STAC catalog will be written to catalog.json in the supplied dir '
                              'with child resources in the same directory or a descendant '
                              'directory.  The remaining message, less any completed operations, '
                              'should be written to message.json in the supplied directory.  If '
                              'there is an error, it will be written to error.json in the supplied dir '))
    parser.add_argument('--harmony-data-location',
                        help=('the location where output data should be written, either a directory '
                              'or S3 URI prefix.  If set, overrides any value set by the message'))
    parser.add_argument('--harmony-queue-url',
                        help='the queue URL to listen on, required for --harmony-action=start')
    parser.add_argument('--harmony-visibility-timeout',
                        type=int,
                        default=600,
                        help=('the number of seconds the service is given to process a message '
                              'before processing is assumed to have failed'))
    parser.add_argument('--harmony-wrap-stdout',
                        action='store_const',
                        const=True,
                        help='Do not wrap STDOUT and STDERR in the Harmony log output format')


def is_harmony_cli(args):
    """
    Returns True if the passed parsed CLI arguments constitute a Harmony CLI invocation, False otherwise

    Parameters
    ----------
    args : Namespace
        Argument values parsed from the command line, presumably via ArgumentParser.parse_args

    Returns
    -------
    is_harmony_cli : bool
        True if the provided arguments constitute a Harmony CLI invocation, False otherwise
    """
    return args.harmony_action is not None


def _invoke_deprecated(AdapterClass, message_string, config):
    """
    Handles --harmony-action=invoke by invoking the adapter for the given input message

    Parameters
    ----------
    AdapterClass : class
        The BaseHarmonyAdapter subclass to use to handle service invocations
    message_string : string
        The Harmony input message
    config : harmony.util.Config
        A configuration instance for this service
    Returns
    -------
    True if the operation completed successfully, False otherwise
    """

    secret_key = config.shared_secret_key
    decrypter = create_decrypter(bytes(secret_key, 'utf-8'))

    message_data = json.loads(message_string)
    adapter = AdapterClass(Message(message_data, decrypter))
    adapter.set_config(config)

    try:
        adapter.invoke()
        if not adapter.is_complete:
            adapter.completed_with_error('The backend service did not respond')

    except CanceledException:
        # If we see the request has been canceled do not try calling back to harmony again
        # Enable this logging after fixing HARMONY-410
        # logging.error('Service request canceled by Harmony, exiting')
        pass
    except HarmonyException as e:
        logging.error(e, exc_info=1)
        if not adapter.is_complete:
            adapter.completed_with_error(str(e))
    except BaseException as e:
        # Make sure we always call back if the error is in a Harmony invocation and we have
        # successfully parsed enough that we know where to call back to
        logging.error(e, exc_info=1)
        if not adapter.is_complete:
            msg = 'Service request failed with an unknown error'
            adapter.completed_with_error(msg)
        raise
    return not adapter.is_failed


def _write_error(metadata_dir, message, category='Unknown'):
    """
    Writes the given error message to error.json in the provided metadata dir

    Parameters
    ----------
    metadata_dir : string
        Directory into which the error should be written
    message : string
        The error message to write
    category : string
        The error category to write
    """
    error_data = {'error': message, 'category': category}
    if is_s3(metadata_dir):
        json_str = json.dumps(error_data)
        write_s3(f'{metadata_dir}error.json', json_str)
    else:
        with open(path.join(metadata_dir, 'error.json'), 'w') as file:
            json.dump(error_data, file)


def _build_adapter(AdapterClass, message_string, sources_path, data_location, config):
    """
    Creates the adapter to be invoked for the given harmony input

    Parameters
    ----------
    AdapterClass : class
        The BaseHarmonyAdapter subclass to use to handle service invocations
    message_string : string
        The Harmony input message
    sources_path : string
        A file location containing a STAC catalog corresponding to the input message sources
    data_location : string
        The name of the directory where output should be written
    config : harmony.util.Config
        A configuration instance for this service
    Returns
    -------
        BaseHarmonyAdapter subclass instance
            The adapter to be invoked
    """
    catalog = Catalog.from_file(sources_path)
    secret_key = config.shared_secret_key

    if bool(secret_key):
        decrypter = create_decrypter(bytes(secret_key, 'utf-8'))
    else:
        def identity(arg):
            return arg
        decrypter = identity

    message = Message(json.loads(message_string), decrypter)
    if data_location:
        message.stagingLocation = data_location
    adapter = AdapterClass(message, catalog)
    adapter.set_config(config)

    return adapter


def _invoke(adapter, metadata_dir):
    """
    Handles --harmony-action=invoke by invoking the adapter for the given input message

    Parameters
    ----------
    adapter : BaseHarmonyAdapter
        The BaseHarmonyAdapter subclass to use to handle service invocations
    metadata_dir : string
        The name of the directory where STAC and message output should be written
    Returns
    -------
    True if the operation completed successfully, False otherwise
    """
    try:
        logging.info(f'Invoking adapter with harmony-service-lib-py version {get_version()}')
        is_s3_metadata_dir = is_s3(metadata_dir)
        if not is_s3_metadata_dir:
            makedirs(metadata_dir, exist_ok=True)
        (out_message, out_catalog) = adapter.invoke()
        out_catalog.normalize_and_save(metadata_dir, CatalogType.SELF_CONTAINED)

        if not is_s3_metadata_dir:
            with open(path.join(metadata_dir, 'message.json'), 'w') as file:
                json.dump(out_message.output_data, file)
    except HarmonyException as err:
        logging.error(err, exc_info=1)
        _write_error(metadata_dir, err.message, err.category)
        raise
    except BaseException as err:
        logging.error(err, exc_info=1)
        _write_error(metadata_dir, 'Service request failed with an unknown error')
        raise


def _start(AdapterClass, queue_url, visibility_timeout_s, config):
    """
    Handles --harmony-action=start by listening to the given queue_url and invoking the
    AdapterClass on any received messages

    Parameters
    ----------
    AdapterClass : class
        The BaseHarmonyAdapter subclass to use to handle service invocations
    queue_url : string
        The SQS queue to listen on
    visibility_timeout_s : int
        The time interval during which the message can't be picked up by other
        listeners on the queue.
    config : harmony.util.Config
        A configuration instance for this service
    """
    for receipt, message in receive_messages(queue_url, visibility_timeout_s, cfg=config):
        # Behavior here is slightly different than _invoke.  Whereas _invoke ensures
        # that the backend receives a callback whenever possible in the case of an
        # exception, the message queue listener prefers to let the message become
        # visibile again and let retry and dead letter queue policies determine visibility
        adapter = AdapterClass(Message(message))
        adapter.set_config(config)

        try:
            adapter.invoke()
        except Exception:
            logging.error('Adapter threw an exception', exc_info=True)
        finally:
            if adapter.is_complete:
                delete_message(queue_url, receipt, cfg=config)
            else:
                change_message_visibility(queue_url, receipt, 0, cfg=config)
            try:
                adapter.cleanup()
            except Exception:
                logging.error(
                    'Adapter threw an exception on cleanup', exc_info=True)


def run_cli(parser, args, AdapterClass, cfg=None):
    """
    Runs the Harmony CLI invocation captured by the given args

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser being used to parse CLI arguments, used to provide CLI argument errors
    args : Namespace
        Argument values parsed from the command line, presumably via ArgumentParser.parse_args
    AdapterClass : class
        The BaseHarmonyAdapter subclass to use to handle service invocations
    cfg : harmony.util.Config
        A configuration instance for this service
    """
    if cfg is None:
        cfg = config()
    if args.harmony_wrap_stdout:
        setup_stdout_log_formatting(cfg)

    if args.harmony_action == 'invoke':
        start_time = datetime.datetime.now()
        if not bool(args.harmony_input):
            parser.error(
                '--harmony-input must be provided for --harmony-action=invoke')
        elif not bool(args.harmony_sources):
            successful = _invoke_deprecated(AdapterClass, args.harmony_input, cfg)
            if not successful:
                raise Exception('Service operation failed')
        else:
            try:
                adapter = _build_adapter(AdapterClass,
                                         args.harmony_input,
                                         args.harmony_sources,
                                         args.harmony_data_location,
                                         cfg)
                adapter.logger.info(f'timing.{cfg.app_name}.start')
                _invoke(adapter, args.harmony_metadata_dir)
            finally:
                time_diff = datetime.datetime.now() - start_time
                duration_ms = int(round(time_diff.total_seconds() * 1000))
                duration_logger = build_logger(cfg)
                extra_fields = {
                    'user': adapter.message.user,
                    'requestId': adapter.message.requestId,
                    'durationMs': duration_ms
                }
                duration_logger.info(f'timing.{cfg.app_name}.end', extra=extra_fields)

    if args.harmony_action == 'start':
        if not bool(args.harmony_queue_url):
            parser.error(
                '--harmony-queue-url must be provided for --harmony-action=start')
        else:
            return _start(AdapterClass, args.harmony_queue_url, args.harmony_visibility_timeout, cfg)



"""
==========
adapter.py
==========

Provides BaseHarmonyAdapter, an abstract base class for services to implement
the translation between Harmony messages and service calls and the translation
between service results and Harmony callbacks.
"""

import shutil
import os
import urllib
import logging
import uuid
from abc import ABC
from tempfile import mkdtemp
from warnings import warn

from deprecation import deprecated
from pystac import Catalog, Item, Asset, read_file

from harmony.exceptions import CanceledException
from harmony.logging import build_logger
from harmony.message import Temporal
from harmony.util import touch_health_check_file
from . import util


class BaseHarmonyAdapter(ABC):
    """
    Abstract base class for Harmony service adapters.  Service implementations
    should inherit from this class and implement the `#invoke(self)` method to
    adapt the Harmony message (`self.message`) into a service call and the
    output of the service call into a response to Harmony (`self.completed_with_*`)

    Services may choose to override methods that do data downloads and result
    staging as well, if they use a different mechanism

    Attributes
    ----------
    message : harmony.Message
        The Harmony input which needs acting upon
    temp_paths : list
        A list of string paths that should be cleaned up on exit
    is_complete : boolean
        True if the service has provided a result to Harmony (and therefore must
        not provide another)
    is_canceled: boolean
        True if the request has been canceled by a Harmony user or operator
    logger: Logger
        Logger specific to this request
    is_failed: boolean
        True if the request failed to execute successfully
    """

    def __init__(self, message, catalog=None, config=None):
        """
        Constructs the adapter

        Parameters
        ----------
        message : harmony.Message
            The Harmony input which needs acting upon
        catalog : pystac.Catalog
            A STAC catalog containing the files on which to act
        config : harmony.util.Config
            The configuration values for this runtime environment.
        """
        if catalog is None:
            warn('Invoking adapter.BaseHarmonyAdapter without a STAC catalog is deprecated',
                 DeprecationWarning, stacklevel=2)

        self.message = message
        self.catalog = catalog
        self.config = config

        if self.config is not None:
            self.init_logging()
        else:
            self.logger = logging.getLogger()

        # Properties that will be deprecated
        self.temp_paths = []
        self.is_complete = False
        self.is_canceled = False
        self.is_failed = False

    def set_config(self, config):
        self.config = config
        if self.config is not None:
            self.init_logging()

    def init_logging(self):
        user = self.message.user if hasattr(self.message, 'user') else None
        req_id = self.message.requestId if hasattr(self.message, 'requestId') else None
        logging_context = {
            'user': user,
            'requestId': req_id
        }
        self.logger = logging.LoggerAdapter(build_logger(self.config), logging_context)

    def invoke(self):
        """
        Invokes the service to process `self.message`.  By default, this will call process_item
        on all items in the input catalog

        Returns
        -------
        (harmony.Message, pystac.Catalog)
            A tuple of the Harmony message, with any processed fields marked as such and
            a STAC catalog describing the output
        """
        # New-style processing using STAC
        if self.catalog:
            return (self.message, self._process_catalog_recursive(self.catalog))

        # Current processing using callbacks
        self._process_with_callbacks()

    def get_all_catalog_items(self, catalog: Catalog, follow_page_links=True):
        """
        Returns a lazy sequence of all the items (including from child catalogs) in the catalog.
        Can handle paged catalogs (catalogs with next/prev).

        Parameters
        ----------
        catalog : pystac.Catalog
            The catalog from which to get items
        follow_page_links : boolean
            Whether or not to follow 'next' links - defaults to True

        Returns
        -------
        A generator that can be iterated to provide a lazy sequence of `pystac.Item`s
        """
        # Return immediate items and items from sub-catalogs
        for item in catalog.get_all_items():
            yield item

        # process 'next' link if present
        if follow_page_links:
            link = catalog.get_single_link(rel='next')
            if link:
                next_catalog = read_file(link.get_href())
                next_items = self.get_all_catalog_items(next_catalog, True)
                for item in next_items:
                    yield item

    def _process_catalog_recursive(self, catalog):
        """
        Helper method to recursively process a catalog and all of its children, producing a new
        output catalog of the results

        Parameters
        ----------
        catalog : pystac.Catalog
            The catalog to process

        Returns
        -------
        pystac.Catalog
            A new catalog containing all of the processed results
        """
        result = catalog.clone()
        result.id = str(uuid.uuid4())

        # Recursively process all sub-catalogs
        children = catalog.get_children()
        result.clear_children()
        result.add_children([self._process_catalog_recursive(child) for child in children])

        # Process immediate child items
        items = catalog.get_items()
        item_count = 0
        result.clear_items()
        source = None
        for item in items:
            item_count = item_count + 1
            source = source or self._get_item_source(item)
            output_item = self.process_item(item.clone(), source)
            if output_item:
                # Ensure the item gets a new ID
                if output_item.id == item.id:
                    output_item.id = str(uuid.uuid4())
                result.add_item(output_item)
        self.logger.info(f'Processed {item_count} granule(s)')

        # process 'next' link if present
        link = catalog.get_single_link(rel='next')
        if link:
            next_catalog = read_file(link.get_href())
            result.add_child(self._process_catalog_recursive(next_catalog))

        return result

    def _process_with_callbacks(self):
        """
        Method for backward compatibility with non-chaining workflows.  Takes an incoming message
        containing granules, translates the granules into STAC items, and passes them individually
        to process_item
        """
        item_count = sum([len(source.granules) for source in self.message.sources])
        completed = 0
        for source in self.message.sources:
            for granule in source.granules:
                item = Item(granule.id, None, granule.bbox, None, {
                    'start_datetime': granule.temporal.start,
                    'end_datetime': granule.temporal.end
                })
                item.add_asset('data', Asset(granule.url, granule.name, roles=['data']))
                result = self.process_item(item, source)
                if not result:
                    continue
                assets = [v for k, v in result.assets.items() if 'data' in (v.roles or [])]
                completed += 1
                progress = int(100 * completed / item_count)
                for asset in assets:
                    temporal = Temporal({}, result.properties['start_datetime'], result.properties['end_datetime'])
                    common_args = dict(
                        title=asset.title,
                        mime=asset.media_type,
                        source_granule=granule,
                        temporal=temporal,
                        bbox=result.bbox
                    )
                    if self.message.isSynchronous:
                        self.completed_with_redirect(asset.href, **common_args)
                        return
                    self.async_add_url_partial_result(asset.href, progress=progress, **common_args)
        self.async_completed_successfully()

    def process_item(self, item, source):
        """
        Given a pystac.Item and a message.Source (collection and variables to subset), processes the
        item, returning a new pystac.Item that describes the output location and metadata

        Optional abstract method. Required if the default #invoke implementation is used.  Services
        processing one input file at a time can simplify adapter code by overriding this method.


        Parameters
        ----------
        item : pystac.Item
            the item that should be processed
        source : harmony.message.Source
            the input source defining the variables, if any, to subset from the item

        Returns
        -------
        pystac.Item
            a STAC item whose metadata and assets describe the service output
        """
        raise NotImplementedError('subclasses must implement #process_item or override #invoke')

    def _get_item_source(self, item):
        """
        Given a STAC item, finds and returns the item's data source in this.message.  It
        specifically looks for a link with relation "harmony_source" in the item and all
        parent catalogs.  The href on that link is the source collection landing page, which
        can identify a source.  If no relation exists and there is only one source in the
        message, returns the message source.

        Parameters
        ----------
        item : pystac.Item
            the item whose source is needed

        Raises
        ------
        RuntimeError
            if no input source could be unambiguously determined, which indiciates a
            misconfiguration or bad input message

        Returns
        -------
        harmony.message.Source
            The source of the input item
        """
        parent = item
        sources = parent.get_links('harmony_source')
        while len(sources) == 0 and parent.get_parent() is not None:
            parent = parent.get_parent()
            sources = parent.get_links('harmony_source')
        if len(sources) == 0:
            if len(self.message.sources) == 1:
                return self.message.sources[0]
            else:
                raise RuntimeError('Could not match STAC catalog to an input source')
        href = sources[0].target
        collection = href.split('/').pop()
        return next(source for source in self.message.sources if source.collection == collection)

    # All methods below are deprecated as we move to STAC-based chaining workflows without callbacks

    @deprecated(details='Services must update to process and output STAC catalogs')
    def cleanup(self):
        """
        Removes temporary files produced during execution
        """
        for temp_path in self.temp_paths:
            if os.path.isfile(temp_path):
                os.remove(temp_path)  # remove the file
            elif os.path.isdir(temp_path):
                shutil.rmtree(temp_path)  # remove dir and all contents
        self.temp_paths = []

    @deprecated(details='Services must update to process and output STAC catalogs')
    def download_granules(self, granules=None):
        """
        Downloads all of the granules contained in the message to the given temp directory, giving each
        a unique filename.

        Parameters
        ----------
        granules : list
            A list of harmony.message.Granule objects corresponding to the granules to download.  Default:
            all granules in the incoming message
        """
        temp_dir = mkdtemp()
        self.temp_paths += [temp_dir]

        granules = granules or self.message.granules

        # Download the remote file
        for granule in granules:
            granule.local_filename = util.download(granule.url, temp_dir, logger=self.logger,
                                                   access_token=self.message.accessToken, cfg=self.config)

    @deprecated(details='Services must update to process and output STAC catalogs')
    def stage(self, local_file, source_granule=None, remote_filename=None, is_variable_subset=False,
              is_regridded=False, is_subsetted=False, mime=None):
        """
        Stages a file on the local filesystem to S3 with the given remote filename and mime type for
        user access.

        Parameters
        ----------
        local_file : string
            The path and name of the file to stage
        source_granule : message.Granule, optional
            The granule from which the file was derived, if it was derived from a single granule.  This
            will be used to produce a canonical filename
        remote_filename : string, optional
            The name of the file when staged, which will be visible to the user requesting data.
            Specify this if not providing a source granule.  If neither remote_filename nor source_granule
            is provided, the output file will use the file's basename
        is_variable_subset : bool, optional
            True if a variable subset operation has been performed (default: False)
        is_regridded : bool, optional
            True if a regridding operation has been performed (default: False)
        is_subsetted : bool, optional
            True if a subsetting operation has been performed (default: False)
        mime : string, optional
            The mime type of the file, by default the output mime type requested by Harmony

        Returns
        -------
        string
            A URI to the staged file
        """
        if remote_filename is None:
            if source_granule:
                remote_filename = self.filename_for_granule(source_granule, os.path.splitext(
                    local_file)[1], is_variable_subset, is_regridded, is_subsetted)
            else:
                remote_filename = os.path.basename(local_file)

        if mime is None:
            mime = self.message.format.mime

        return util.stage(local_file, remote_filename, mime, location=self.message.stagingLocation,
                          logger=self.logger, cfg=self.config)

    @deprecated(details='Services must update to process and output STAC catalogs')
    def completed_with_error(self, error_message):
        """
        Performs a callback instructing Harmony that there has been an error and providing a
        message to send back to the service user

        Parameters
        ----------
        error_message : string
            The error message to pass on to the service user

        Raises
        ------
        Exception
            If a callback has already been performed
        """
        self.is_failed = True
        if self.is_complete and not self.is_canceled:
            raise Exception(
                'Attempted to error an already-complete service call with message ' + error_message)
        self._callback_response({'error': error_message})
        self.is_complete = True

    @deprecated(details='Services must update to process and output STAC catalogs')
    def completed_with_redirect(
            self,
            url,
            title=None,
            mime=None,
            source_granule=None,
            temporal=None,
            bbox=None):
        """
        Performs a callback instructing Harmony to redirect the service user to the given URL

        Parameters
        ----------
        url : string
            The URL where the service user should be redirected
        mime : string, optional
            The mime type of the file, by default the output mime type requested by Harmony
        title : string, optional
            Textual information to provide users along with the link
        temporal : harmony.message.Temporal, optional
            The temporal extent of the provided file.  If not provided, the source granule's
            temporal will be used when a source granule is provided
        bbox : list, optional
            List of [West, South, East, North] for the MBR of the provided result.  If not provided,
            the source granule's bbox will be used when a source granule is provided

        Raises
        ------
        Exception
            If a callback has already been performed
        """

        if self.is_complete:
            raise Exception(
                'Attempted to redirect an already-complete service call to ' + url)
        params = self._build_callback_item_params(url, mime=mime, source_granule=source_granule)
        params['status'] = 'successful'
        self._callback_response(params)
        self.is_complete = True

    @deprecated(details='Services must update to process and output STAC catalogs')
    def completed_with_local_file(
            self,
            filename,
            source_granule=None,
            remote_filename=None,
            is_variable_subset=False,
            is_regridded=False,
            is_subsetted=False,
            mime=None,
            title=None,
            temporal=None,
            bbox=None):
        """
        Indicates that the service has completed with the given file as its result.  Stages the
        provided local file to a user-accessible S3 location and instructs Harmony to redirect
        to that location.

        Parameters
        ----------
        filename : string
            The path and name of the local file
        source_granule : message.Granule, optional
            The granule from which the file was derived, if it was derived from a single granule.  This
            will be used to produce a canonical filename
        remote_filename : string, optional
            The name of the file when staged, which will be visible to the user requesting data.
            Specify this if not providing a source granule.  If neither remote_filename nor source_granule
            is provided, the output file will use the file's basename
        is_variable_subset : bool, optional
            True if a variable subset operation has been performed (default: False)
        is_regridded : bool, optional
            True if a regridding operation has been performed (default: False)
        is_subsetted : bool, optional
            True if a subsetting operation has been performed (default: False)
        mime : string, optional
            The mime type of the file, by default the output mime type requested by Harmony
        title : string, optional
            Textual information to provide users along with the link
        temporal : harmony.message.Temporal, optional
            The temporal extent of the provided file.  If not provided, the source granule's
            temporal will be used when a source granule is provided
        bbox : list, optional
            List of [West, South, East, North] for the MBR of the provided result.  If not provided,
            the source granule's bbox will be used when a source granule is provided

        Raises
        ------
        Exception
            If a callback has already been performed
        """
        url = self.stage(filename, source_granule, remote_filename,
                         is_variable_subset, is_regridded, is_subsetted, mime)
        self.completed_with_redirect(url, title, mime, source_granule, temporal, bbox)

    @deprecated(details='Services must update to process and output STAC catalogs')
    def async_add_local_file_partial_result(
            self,
            filename,
            source_granule=None,
            remote_filename=None,
            is_variable_subset=False,
            is_regridded=False,
            is_subsetted=False,
            title=None,
            mime=None,
            progress=None,
            temporal=None,
            bbox=None):
        """
        For service requests that are asynchronous, stages the given filename and sends the staged
        URL as a progress update to Harmony.  Optionally also provides a numeric progress indicator.
        Synchronous requests may not call this method and will throw an exeception.

        Parameters
        ----------
        filename : string
            The path and name of the local file
        source_granule : message.Granule, optional
            The granule from which the file was derived, if it was derived from a single granule.  This
            will be used to produce a canonical filename and assist when temporal and bbox are not specified
        remote_filename : string, optional
            The name of the file when staged, which will be visible to the user requesting data.
            Specify this if not providing a source granule.  If neither remote_filename nor source_granule
            is provided, the output file will use the file's basename
        is_variable_subset : bool, optional
            True if a variable subset operation has been performed (default: False)
        is_regridded : bool, optional
            True if a regridding operation has been performed (default: False)
        is_subsetted : bool, optional
            True if a subsetting operation has been performed (default: False)
        title : string, optional
            Textual information to provide users along with the link
        mime : string, optional
            The mime type of the file, by default the output mime type requested by Harmony
        progress : integer, optional
            Numeric progress of the total request, 0-100
        temporal : harmony.message.Temporal, optional
            The temporal extent of the provided file.  If not provided, the source granule's
            temporal will be used when a source granule is provided
        bbox : list, optional
            List of [West, South, East, North] for the MBR of the provided result.  If not provided,
            the source granule's bbox will be used when a source granule is provided

        Raises
        ------
        Exception
            If the request is synchronous or the request has already been marked complete
        """
        url = self.stage(filename, source_granule, remote_filename,
                         is_variable_subset, is_regridded, is_subsetted, mime)
        self.async_add_url_partial_result(url, title, mime, progress, source_granule,
                                          temporal, bbox)

    @deprecated(details='Services must update to process and output STAC catalogs')
    def async_add_url_partial_result(self, url, title=None, mime=None, progress=None, source_granule=None,
                                     temporal=None, bbox=None):
        """
        For service requests that are asynchronous, stages the provides the given URL as a partial result.
        Optionally also provides a numeric progress indicator.
        Synchronous requests may not call this method and will throw an exeception.

        Parameters
        ----------
        url : string
            The URL where the service user should be redirected
        title : string, optional
            Textual information to provide users along with the link
        mime : string, optional
            The mime type of the file, by default the output mime type requested by Harmony
        progress : integer, optional
            Numeric progress of the total request, 0-100
        source_granule : message.Granule, optional
            The granule from which the file was derived, if it was derived from a single granule.  This
            will be used to produce a canonical filename and assist when temporal and bbox are not specified
        temporal : harmony.message.Temporal, optional
            The temporal extent of the provided file.  If not provided, the source granule's
            temporal will be used when a source granule is provided
        bbox : list, optional
            List of [West, South, East, North] for the MBR of the provided result.  If not provided,
            the source granule's bbox will be used when a source granule is provided

        Raises
        ------
        Exception
            If the request is synchronous or the request has already been marked complete
        """
        if self.message.isSynchronous:
            raise Exception(
                'Attempted to call back asynchronously to a synchronous request')
        if self.is_complete:
            raise Exception(
                'Attempted to add a result to an already-completed request: ' + url)
        params = self._build_callback_item_params(url, title, mime, source_granule, temporal, bbox)
        if progress is not None:
            params['progress'] = progress

        self._callback_response(params)

    @deprecated(details='Services must update to process and output STAC catalogs')
    def async_completed_successfully(self):
        """
        For service requests that are asynchronous, sends a progress update indicating
        that a service request is complete.
        Synchronous requests may not call this method and will throw an exeception.

        Raises
        ------
        Exception
            If the request is synchronous or the request has already been marked complete
        """
        if self.message.isSynchronous:
            raise Exception(
                'Attempted to call back asynchronously to a synchronous request')
        if self.is_complete:
            raise Exception(
                'Attempted to call back for an already-completed request.')
        self._callback_response({'status': 'successful'})
        self.is_complete = True

    @deprecated(details='Services must update to process and output STAC catalogs')
    def filename_for_granule(self, granule, ext, is_variable_subset=False, is_regridded=False, is_subsetted=False):
        """
        Return an output filename for the given granules according to our naming conventions:
        {original filename without suffix}(_{single var})?(_regridded)?(_subsetted)?.<ext>

        Parameters
        ----------
            granule : message.Granule
                The source granule for the output file
            ext: string
                The destination file extension
            is_variable_subset : bool, optional
                True if a variable subset operation has been performed (default: False)
            is_regridded : bool, optional
                True if a regridding operation has been performed (default: False)
            is_subsetted : bool, optional
                True if a subsetting operation has been performed (default: False)

        Returns
        -------
            string
                The output filename
        """
        url = granule.url
        # Get everything between the last non-trailing '/' before the query and the first '?'
        # Do this instead of using a URL parser, because our URLs are not complex in practice and
        # it is useful to allow relative file paths to work for local testing.
        original_filename = url.split('?')[0].rstrip('/').split('/')[-1]
        original_basename = os.path.splitext(original_filename)[0]
        if not ext.startswith('.'):
            ext = '.' + ext

        suffixes = []
        if is_variable_subset and len(granule.variables) == 1:
            suffixes.append('_' + granule.variables[0].name.replace('/', '_'))
        if is_regridded:
            suffixes.append('_regridded')
        if is_subsetted:
            suffixes.append('_subsetted')
        suffixes.append(ext)

        result = original_basename
        # Iterate suffixes in reverse, removing them from the result if they're at the end of the string
        # This supports the case of chaining where one service regrids and another subsets but we don't
        # want names to get mangled
        for suffix in suffixes[::-1]:
            if result.endswith(suffix):
                result = result[:-len(suffix)]

        return result + "".join(suffixes)

    # Deprecated internal methods below

    def _build_callback_item_params(
            self,
            url,
            title=None,
            mime=None,
            source_granule=None,
            temporal=None,
            bbox=None):
        """
        Builds the "item[...]" parameters required for a callback to Harmony for the given
        params, returning them as a string param / string value dict.

        Parameters
        ----------
        url : string
            The URL where the service user should be redirected
        title : string, optional
            Textual information to provide users along with the link
        mime : string, optional
            The mime type of the file, by default the output mime type requested by Harmony
        source_granule : message.Granule, optional
            The granule from which the file was derived, if it was derived from a single granule.  This
            will be used to produce a canonical filename and assist when temporal and bbox are not specified
        temporal : harmony.message.Temporal, optional
            The temporal extent of the provided file.  If not provided, the source granule's
            temporal will be used when a source granule is provided
        bbox : list, optional
            List of [West, South, East, North] for the MBR of the provided result.  If not provided,
            the source granule's bbox will be used when a source granule is provided

        Returns
        -------
        dict
            A dictionary containing a mapping of query parameters to value for the given params
        """
        if mime is None:
            mime = self.message.format.mime
        if source_granule is not None:
            temporal = temporal or source_granule.temporal
            bbox = bbox or source_granule.bbox

        params = {'item[href]': url, 'item[type]': mime}
        if title is not None:
            params['item[title]'] = title
        if temporal is not None:
            params['item[temporal]'] = ','.join([temporal.start, temporal.end])
        if bbox is not None:
            params['item[bbox]'] = ','.join([str(c) for c in bbox])
        return params

    def _callback_response(self, query_params):
        """
        POSTs to the Harmony callback URL at the given path with the given params

        Parameters
        ----------
        query_params : dict
            A mapping of string key to string value query params to send to the callback

        Returns
        -------
        None
        """

        param_strs = ['%s=%s' % (k, urllib.parse.quote(str(v)))
                      for k, v in query_params.items()]
        self._callback_post('/response?' + '&'.join(param_strs))

    def _callback_post(self, path):
        """
        POSTs to the Harmony callback URL at the given path, which may include query params

        Parameters
        ----------
        path : string
            The URL path relative to the Harmony callback URL which should be POSTed to

        Returns
        -------
        None
        """

        url = self.message.callback + path
        touch_health_check_file(self.config.health_check_path)
        if os.environ.get('ENV') in ['dev', 'test']:
            self.logger.warning(
                'ENV=' + os.environ['ENV'] + ' so we will not reply to Harmony with POST ' + url)
        elif self.is_canceled:
            msg = 'Ignoring making callback request because the request has been canceled.'
            self.logger.info(msg)
        else:
            self.logger.info('Starting response: %s', url)
            request = urllib.request.Request(url, method='POST')
            try:
                response = \
                    urllib.request.urlopen(request).read().decode('utf-8')
                self.logger.info('Remote response: %s', response)
                self.logger.info('Completed response: %s', url)
            except urllib.error.HTTPError as e:
                self.is_failed = True
                body = e.read().decode()
                msg = f'Harmony returned an error when updating the job: {body}'
                self.logger.error(msg, exc_info=e)
                if e.code == 409:
                    self.logger.warning('Harmony request was canceled.')
                    self.is_canceled = True
                    self.is_complete = True
                    raise CanceledException
                raise
            except Exception as e:
                self.is_failed = True
                self.logger.error('Error when updating the job', exc_info=e)
                raise



import os
from tempfile import mkdtemp
import shutil
import unittest

from pystac import Catalog, CatalogType

from harmony import cli, BaseHarmonyAdapter
from harmony.exceptions import ForbiddenException
from tests.util import cli_parser, config_fixture


class MockAdapter(BaseHarmonyAdapter):
    message = None
    """
    Dummy class to mock adapter calls, performing a no-op service
    """
    def invoke(self):
        MockAdapter.message = self.message
        return (self.message, self.catalog)


class TestCliInvokeAction(unittest.TestCase):
    def setUp(self):
        self.workdir = mkdtemp()
        self.inputdir = mkdtemp()
        self.catalog = Catalog('test-id', 'test catalog')
        self.catalog.normalize_and_save(self.inputdir, CatalogType.SELF_CONTAINED)
        self.config = config_fixture()
        print(self.config)

    def tearDown(self):
        MockAdapter.messages = []
        shutil.rmtree(self.workdir)

    def test_when_a_service_completes_it_writes_a_output_catalog_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, cfg=self.config)
            output = Catalog.from_file(os.path.join(self.workdir, 'catalog.json'))
            self.assertTrue(output.validate)

    def test_when_a_service_completes_it_writes_the_output_message_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, cfg=self.config)
            with open(os.path.join(self.workdir, 'message.json')) as file:
                self.assertEqual(file.read(), '{"test": "input"}')

    def test_when_the_cli_has_a_staging_location_it_overwites_the_message_staging_location(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir,
                '--harmony-data-location', 's3://fake-location/') as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, cfg=self.config)
            self.assertEqual(MockAdapter.message.stagingLocation, 's3://fake-location/')
            # Does not output the altered staging location
            with open(os.path.join(self.workdir, 'message.json')) as file:
                self.assertEqual(file.read(), '{"test": "input"}')

    def test_when_the_backend_service_throws_a_known_error_it_writes_the_error_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:

            class MockImpl(MockAdapter):
                def invoke(self):
                    self.is_complete = False
                    raise ForbiddenException('Something bad happened')

            args = parser.parse_args()
            with self.assertRaises(Exception) as context:
                cli.run_cli(parser, args, MockImpl, cfg=self.config)

            self.assertTrue('Something bad happened' in str(context.exception))
            with open(os.path.join(self.workdir, 'error.json')) as file:
                self.assertEqual(file.read(), '{"error": "Something bad happened", "category": "Forbidden"}')

    def test_when_the_backend_service_throws_an_unknown_error_it_writes_a_generic_error_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:

            class MockImpl(MockAdapter):
                def invoke(self):
                    self.is_complete = False
                    raise Exception('Something bad happened')

            args = parser.parse_args()
            with self.assertRaises(Exception) as context:
                cli.run_cli(parser, args, MockImpl, cfg=self.config)

            self.assertTrue('Something bad happened' in str(context.exception))
            with open(os.path.join(self.workdir, 'error.json')) as file:
                self.assertEqual(file.read(), '{"error": "Service request failed with an unknown error", "category": "Unknown"}')


if __name__ == '__main__':
    unittest.main()



from base64 import b64encode
import re
from urllib.parse import urlparse

from requests.auth import AuthBase
from requests import Session

EDL_URL_PATTERN = r""".*urs\.earthdata\.nasa\.gov$"""


def _edl_url(url: str) -> bool:
    """Determine if the given URL is for Earthdata Login."""
    hostname = urlparse(url).hostname
    return re.fullmatch(EDL_URL_PATTERN, hostname) is not None


class EarthdataSession(Session):
    """Session which ensures the Authorization header is sent to correct
    servers.

    After instantiating the EarthdataSession, set its `auth` attribute
    to a valid EarthdataAuth instance:

        session.auth = EarthdataAuth(...)

    This lifecycle method on requests.Session is called when handling
    redirect requests. There are two cases important for handling
    Earthdata Login:

    (A) When handling a redirect from a resource server to Earthdata
    Login, the session will use the auth (if provided) to add the
    required Authorization to the request.

    (B) When handling a redirect from Earthdata Login back to a
    resource server, the session will remove the Authorization header
    from the request (which the `requests` package copies from the
    request which caused this redirect).

    """
    def rebuild_auth(self, prepared_request, response):
        # If not configured with an EarthdataAuth instance, defer to
        # default behavior
        if not self.auth:
            return super().rebuild_auth(prepared_request, response)

        if _edl_url(prepared_request.url):
            # (A) Defer to auth to add the Authorization header
            self.auth(prepared_request)
        else:
            # (B) Remove the Authorization header when redirecting away
            # from EDL.
            prepared_request.headers.pop('Authorization', None)


class EarthdataAuth(AuthBase):
    """Custom Earthdata Auth provider to add EDL Authorization headers to
    requests when required for token sharing and federated
    authentication.

    When instantiated with an EDL application's credentials and a
    user's access token, the resulting HTTP Authorization header will
    include the properly-encoded app credentials as a Basic auth
    header, and the user's access token as a Bearer auth header.

    """
    def __init__(self, app_uid: str, app_pwd: str, user_access_token: str):
        """Instantiate the Earthdata Auth provider.

        Parameters
        ----------
        app_uid:
            The Earthdata Login Application `uid`.

        app_pwd:
            The Earthdata Login Application `password`.

        user_access_token:
            The EDL-issued token for the user making the request.
        """
        creds = b64encode(f"{app_uid}:{app_pwd}".encode('utf-8')).decode('utf-8')
        self.authorization_header = f'Basic {creds}, Bearer {user_access_token}'

    def __call__(self, r):
        """The EarthdataAuth is a callable which adds Authorization headers
        when handling a request for Earthdata Login.

        """
        if _edl_url(r.url):
            r.headers['Authorization'] = self.authorization_header
        return r



import unittest
from unittest.mock import patch

import harmony.util
from harmony import cli, BaseHarmonyAdapter
from tests.util import mock_receive, cli_test


class MockAdapter(BaseHarmonyAdapter):
    """
    Dummy class to mock adapter calls and record the input messages
    """
    messages = []
    errors = []
    cleaned_up = []

    def __init__(self, message):
        super().__init__(self, message)
        MockAdapter.messages.append(message.data)

    def invoke(self):
        self.is_complete = True
        self.is_failed = False

    def completed_with_error(self, error):
        MockAdapter.errors.append(error)

    def cleanup(self):
        MockAdapter.cleaned_up.append(True)


class TestIsHarmonyCli(unittest.TestCase):
    @cli_test('--something-else', 'invoke')
    def test_when_not_passing_harmony_action_it_returns_false(self, parser):
        parser.add_argument('--something-else')
        args = parser.parse_args()
        self.assertFalse(cli.is_harmony_cli(args))

    @cli_test('--harmony-action', 'invoke')
    def test_when_passing_harmony_action_it_returns_true(self, parser):
        args = parser.parse_args()
        self.assertTrue(cli.is_harmony_cli(args))

    @cli_test()
    def test_when_passing_nothing_it_returns_false(self, parser):
        args = parser.parse_args()
        self.assertFalse(cli.is_harmony_cli(args))


class TestCliInvokeAction(unittest.TestCase):
    def setUp(self):
        self.config = harmony.util.config(validate=False)

    def tearDown(self):
        MockAdapter.messages = []
        MockAdapter.errors = []
        MockAdapter.cleaned_up = []

    @cli_test('--harmony-action', 'invoke')
    def test_when_harmony_input_is_not_provided_it_terminates_with_error(self, parser):
        with patch.object(parser, 'error') as error_method:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, self.config)
            error_method.assert_called_once_with(
                '--harmony-input must be provided for --harmony-action=invoke')

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}')
    def test_when_harmony_input_is_provided_it_creates_and_invokes_an_adapter(self, parser):
        args = parser.parse_args()
        cli.run_cli(parser, args, MockAdapter, self.config)
        self.assertListEqual([{'test': 'input'}], MockAdapter.messages)

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}')
    def test_when_the_backend_service_doesnt_respond_it_responds_with_an_error(self, parser):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = False

        args = parser.parse_args()
        try:
            cli.run_cli(parser, args, MockImpl, self.config)
        except Exception:
            pass
        self.assertListEqual(
            MockImpl.errors, ['The backend service did not respond'])

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}')
    def test_when_the_backend_service_throws_an_exception_before_response_it_responds_with_an_error(self, parser):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = False
                raise Exception('Something bad happened')

        args = parser.parse_args()
        with self.assertRaises(Exception) as context:
            cli.run_cli(parser, args, MockImpl, self.config)

        self.assertTrue('Something bad happened' in str(context.exception))
        self.assertListEqual(
            MockImpl.errors, ['Service request failed with an unknown error'])

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}')
    def test_when_the_backend_service_throws_an_exception_afterresponse_it_does_not_respond_again(self, parser):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = True
                raise Exception('Something bad happened')

        args = parser.parse_args()
        try:
            cli.run_cli(parser, args, MockImpl, self.config)
        except Exception:
            pass
        self.assertListEqual(MockImpl.errors, [])


class TestCliStartAction(unittest.TestCase):
    def setUp(self):
        self.config = harmony.util.config(validate=False)

    def tearDown(self):
        MockAdapter.messages = []
        MockAdapter.errors = []
        MockAdapter.cleaned_up = []

    @cli_test('--harmony-action', 'start')
    def test_when_queue_url_is_not_provided_it_terminates_with_error(self, parser):
        with patch.object(parser, 'error') as error_method:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, self.config)
            error_method.assert_called_once_with(
                '--harmony-queue-url must be provided for --harmony-action=start')

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_listens_on_the_provided_queue(self, parser, client):
        sqs = mock_receive(self.config, client, parser, MockAdapter)
        sqs.receive_message.assert_called_with(
            QueueUrl='test-queue-url',
            VisibilityTimeout=600,
            WaitTimeSeconds=20,
            MaxNumberOfMessages=1)
        self.assertListEqual(MockAdapter.messages, [])
        self.assertListEqual(MockAdapter.errors, [])

    @cli_test('--harmony-action', 'start',
              '--harmony-queue-url', 'test-queue-url',
              '--harmony-visibility-timeout', '100')
    @patch('boto3.client')
    def test_uses_optional_visibility_timeouts_from_the_command_line(self, parser, client):
        sqs = mock_receive(self.config, client, parser, MockAdapter)
        sqs.receive_message.assert_called_with(
            QueueUrl='test-queue-url',
            VisibilityTimeout=100,
            WaitTimeSeconds=20,
            MaxNumberOfMessages=1)
        self.assertListEqual(MockAdapter.messages, [])
        self.assertListEqual(MockAdapter.errors, [])

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_sends_queue_messages_to_the_adapter(self, parser, client):
        mock_receive(self.config, client, parser, MockAdapter,
                     '{"test": "a"}', None, '{"test": "b"}')
        self.assertEqual(MockAdapter.messages, [{'test': 'a'}, {'test': 'b'}])

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_completes_the_request_it_deletes_the_queue_message(self, parser, client):
        sqs = mock_receive(self.config, client, parser, MockAdapter,
                           '{"test": "a"}', None, '{"test": "b"}')
        sqs.delete_message.assert_called_with(
            QueueUrl='test-queue-url',
            ReceiptHandle=2)

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_completes_the_request_it_calls_cleanup_on_the_adapter(self, parser, client):
        mock_receive(self.config, client, parser, MockAdapter,
                     '{"test": "a"}', None, '{"test": "b"}')
        self.assertListEqual(MockAdapter.cleaned_up, [True, True])

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_runs_without_completing_the_request_it_returns_the_message_to_the_queue(
        self, parser, client
    ):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = False

        sqs = mock_receive(self.config, client, parser, MockImpl,
                           '{"test": "a"}', None, '{"test": "b"}')
        sqs.delete_message.assert_not_called()
        sqs.change_message_visibility.assert_called_with(
            QueueUrl='test-queue-url',
            VisibilityTimeout=0,
            ReceiptHandle=2)

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_runs_without_completing_the_request_it_calls_cleanup_on_the_adapter(self, parser, client):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = False

        mock_receive(self.config, client, parser, MockImpl, '{"test": "a"}')
        self.assertListEqual(MockImpl.cleaned_up, [True])

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_throws_before_completing_the_request_it_returns_the_message_to_the_queue(
        self, parser, client
    ):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = False
                raise Exception('Something bad happened')

        sqs = mock_receive(self.config, client, parser, MockImpl,
                           '{"test": "a"}', None, '{"test": "b"}')
        sqs.delete_message.assert_not_called()
        sqs.change_message_visibility.assert_called_with(
            QueueUrl='test-queue-url',
            VisibilityTimeout=0,
            ReceiptHandle=2)

        self.assertListEqual(MockImpl.cleaned_up, [True, True])

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_throws_after_completing_the_request_it_deletes_the_queue_message(self, parser, client):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = True
                raise Exception('Something bad happened')

        sqs = mock_receive(self.config, client, parser, MockImpl,
                           '{"test": "a"}', None, '{"test": "b"}')
        sqs.delete_message.assert_called_with(
            QueueUrl='test-queue-url',
            ReceiptHandle=2)
        sqs.change_message_visibility.assert_not_called()

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_the_adapter_throws_it_calls_cleanup_on_the_adapter(self, parser, client):
        class MockImpl(MockAdapter):
            def invoke(self):
                self.is_complete = False
                raise Exception('Something bad happened')

        mock_receive(self.config, client, parser, MockImpl, '{"test": "a"}', None, '{"test": "b"}')
        self.assertListEqual(MockImpl.cleaned_up, [True, True])

    @cli_test('--harmony-action', 'start', '--harmony-queue-url', 'test-queue-url')
    @patch('boto3.client')
    def test_when_cleanup_throws_it_continues_processing_queue_messages(self, parser, client):
        class MockImpl(MockAdapter):
            def cleanup(self):
                raise Exception('Something bad happened')

        mock_receive(self.config, client, parser, MockImpl, '{"test": "a"}', None, '{"test": "b"}')
        self.assertEqual(MockAdapter.messages, [{'test': 'a'}, {'test': 'b'}])


if __name__ == '__main__':
    unittest.main()



"""
===========
__init__.py
===========

Convenience exports for the Harmony library
"""

# Automatically updated by `make build`
__version__ = "v1.0.23"

from .adapter import BaseHarmonyAdapter
from .cli import setup_cli, is_harmony_cli, run_cli
from .message import Temporal
from pystac.stac_io import STAC_IO
from .s3_stac_io import read, write

STAC_IO.read_text_method = read
STAC_IO.write_text_method = write



from urllib.parse import urlparse
import boto3
from pystac import STAC_IO
from harmony import util
from harmony import aws
from os import environ

"""
Read and write to s3 when STAC links start with s3://.
https://pystac.readthedocs.io/en/0.5/concepts.html#using-stac-io
"""


def read(uri):
    """
    Reads STAC files from s3
    (or via the default method if the protocol is not s3).

    Parameters
    ----------
    uri: The STAC file uri.

    Returns
    -------
    The file contents
    """
    config = util.config(validate=environ.get('ENV') != 'test')
    service_params = aws.aws_parameters(
        config.use_localstack, config.localstack_host, config.aws_default_region)
    parsed = urlparse(uri)
    if parsed.scheme == 's3':
        bucket = parsed.netloc
        key = parsed.path[1:]
        s3 = boto3.resource('s3', **service_params)
        obj = s3.Object(bucket, key)
        return obj.get()['Body'].read().decode('utf-8')
    else:
        return STAC_IO.default_read_text_method(uri)


def write(uri, txt):
    """
    Writes a STAC file to the given uri.

    Parameters
    ----------
    uri: The STAC file uri.
    txt: The STAC contents.
    """
    parsed = urlparse(uri)
    if parsed.scheme == 's3':
        aws.write_s3(uri, txt)
    else:
        STAC_IO.default_write_text_method(uri, txt)



from base64 import b64encode
from dataclasses import dataclass, field

import pytest
from requests import Session

from harmony.earthdata import EarthdataAuth, EarthdataSession


@dataclass
class FakeRequest:
    url: str = 'https://uat.urs.earthdata.nasa.gov/oauth'
    headers: dict = field(default_factory=dict)


@pytest.fixture
def earthdata_auth(faker):
    uid = faker.simple_profile('username')
    pwd = faker.password(length=12, special_chars=False)
    token = faker.password(length=40, special_chars=False)
    return EarthdataAuth(uid, pwd, token)


def test_authdata_auth_creates_correct_header(faker):
    uid = faker.simple_profile('username')
    pwd = faker.password(length=12, special_chars=False)
    token = faker.password(length=40, special_chars=False)
    auth = EarthdataAuth(uid, pwd, token)
    request = FakeRequest()

    auth(request)

    assert 'Authorization' in request.headers
    expected_creds = b64encode(f'{uid}:{pwd}'.encode('utf-8')).decode('utf-8')
    assert 'Basic' in request.headers['Authorization']
    assert expected_creds in request.headers['Authorization']
    assert 'Bearer' in request.headers['Authorization']
    assert token in request.headers['Authorization']


def test_earthdata_auth_given_edl_url_adds_auth_header(earthdata_auth):
    request = FakeRequest()

    earthdata_auth(request)

    assert 'Authorization' in request.headers


def test_earthdata_auth_given_non_edl_url_does_not_add_header(earthdata_auth):
    request = FakeRequest()
    request.url = 'https://github.com/acme/foobar'

    earthdata_auth(request)

    assert 'Authorization' not in request.headers


# Expected EarthdataSession behavior:
#
# Auth: False => delegates to super()
# Auth: True  =>
#       | Request Has Header | Request for EDL | Auth Header? |
#       |       False        |      False      |    False     |
#       |       False        |      True       |    True      |
#       |       True         |      False      |    False     |
#       |       True         |      True       |    True (*)  |
# Note:
#   (*) Replace pre-existing Authorization header with new header
#

def test_earthdata_session_given_no_auth_delegates_to_super(monkeypatch):
    called = False

    def mock_rebuild_auth(self, prepared_request, response):
        nonlocal called
        called = True
    monkeypatch.setattr(Session, 'rebuild_auth', mock_rebuild_auth)
    session = EarthdataSession()

    session.rebuild_auth(None, None)

    assert called


def test_earthdata_session_given_no_header_and_non_edl_url_request_does_not_contain_header(earthdata_auth):
    session = EarthdataSession()
    session.auth = earthdata_auth
    request = FakeRequest('https://duckduckgo.com/')

    session.rebuild_auth(request, None)

    assert 'Authorization' not in request.headers


def test_earthdata_session_given_no_header_and_edl_url_request_contains_new_header(earthdata_auth):
    session = EarthdataSession()
    session.auth = earthdata_auth
    request = FakeRequest()

    session.rebuild_auth(request, None)

    assert 'Authorization' in request.headers


def test_earthdata_session_given_header_and_non_edl_url_request_does_not_contain_header(earthdata_auth):
    session = EarthdataSession()
    session.auth = earthdata_auth
    request = FakeRequest('https://duckduckgo.com/', {'Authorization': 'PreExistingValue'})

    session.rebuild_auth(request, None)

    assert 'Authorization' not in request.headers


def test_earthdata_session_given_header_and_edl_url_request_contains_existing_header(earthdata_auth):
    session = EarthdataSession()
    session.auth = earthdata_auth
    request = FakeRequest(headers={'Authorization': 'PreExistingValue'})

    session.rebuild_auth(request, None)

    assert 'Authorization' in request.headers
    assert request.headers['Authorization'] != 'PreExistingValue'
